from typing import Any

from pydantic                import ValidationError
from werkzeug.datastructures import MultiDict
from werkzeug.exceptions     import BadRequest

from nsj_flask_utils.dtos.base_dto import BaseDTO
from nsj_flask_utils.exceptions    import HTTP400

from .filter_params     import FilterParams
from .pagination_params import PaginationParams, PaginationDTO

class RequestArgs:
    filter_params: FilterParams
    pagi_params  : PaginationParams

    def __init__(self, dto_cls: type[BaseDTO]
                     , args   : MultiDict[str, str]) -> None:
        (obj_fields, pagi_dto) = self._validate(dto_cls, args)
        self.filter_params = FilterParams(obj_fields)
        self.pagi_params   = PaginationParams(
            dto_cls, pagi_dto
        )
        pass

    def _validate(self, dto_cls: type[BaseDTO]
                      , args: MultiDict[str, str]) -> tuple[dict[str, Any], PaginationDTO]:
        obj_fields_name : list[str] = list(dto_cls.model_fields.keys())
        pagi_fields_name: list[str] = list(PaginationDTO.model_fields.keys())

        # Possible Fields
        p_fields: list[str] = []
        p_fields.extend(obj_fields_name)
        p_fields.extend(pagi_fields_name)

        for k in args.keys():
            if not k in p_fields:
                if k != 'tenant':
                    raise BadRequest(f"Field {k} isn't a valid field.")
            pass

        obj_fields: dict[str, Any] = {
            k: v
            for k, v in args.items()
            if k in obj_fields_name
        }

        try:
            dto_cls.validate(obj_fields)
        except ValidationError as e:
            errs = {
                x['loc'][0]: x['msg']
                for x in e.errors()
            }
            matchs: list[str] = [
                x
                for x in args.keys()
                if x in errs
            ]

            if matchs:
                errdict = {
                    k: v
                    for k, v in errs.items()
                    if k in matchs
                }
                raise HTTP400({
                    'message': 'Invalid filter params',
                    'erros': errdict,
                })

        pagi_fields: dict[str, Any] = {
            k: v
            for k, v in args.items()
            if k in pagi_fields_name
        }

        pagi_dto: PaginationDTO
        try:
            pagi_dto = PaginationDTO(**pagi_fields)
        except ValidationError as e:
            errs = {
                x['loc'][0]: x['msg']
                for x in e.errors()
            }
            raise HTTP400({
                'message': 'Invalid filter params',
                'erros': errs
            })

        return obj_fields, pagi_dto

    def get_filter_params(self) -> dict[str, str]:
        return self.filter_params.__dict__

    def get_pagination_params(self) -> dict[str, str | int]:
        return self.pagi_params.to_dict()



from typing import ClassVar

from pydantic import BaseModel

class BaseDTO(BaseModel):
    __order_by__  : ClassVar
    __table_name__: ClassVar
    __schema__    : ClassVar
    __PK__        : ClassVar
    pass



from typing import Generic, Any, Mapping, Generic, TypeVar

import sqlalchemy
from werkzeug.exceptions import NotFound
from sqlalchemy.engine   import Connection, CursorResult, Transaction

from nsj_flask_utils.dtos.base_dto import BaseDTO

TDTO = TypeVar('TDTO', bound=BaseDTO)

class BaseDAO(Generic[TDTO]):
    _db : Connection
    _dto: type[TDTO]

    def __init__(self, db : Connection
                     , dto: type[TDTO] | None = None) -> None:
        self._db = db
        if dto:
            self._dto = dto
            pass
        pass

    def _execute(self, query: str
                     , params: Mapping[str, Any] = {}) -> CursorResult[Any]:
        tran: Transaction = self._db.begin()
        rs = self._db.execute(sqlalchemy.text(query), params)
        tran.commit()
        return rs

    def _get_ftbl_name(self) -> str:
        """Returns the table name formatede: `schema.tbl_name'"""
        return f"{self._dto.__schema__}.{self._dto.__table_name__}"


    def _list(self, query: str, pagi_args  : dict[str, str | int]
                              , filter_args: dict[str, str]) -> list[TDTO]:
        query = self._filter_query(query, filter_args)

        query += f"""
ORDER BY {pagi_args['order'],} {pagi_args['order_by'],}
LIMIT :limit
OFFSET :offset
"""

        params = {
            **pagi_args,
            **filter_args,
        }

        rs = self._execute(query, params)
        return [
            self._dto(**x)
            for x in rs.mappings().all()
        ]

    def _get(self, query: str, params: dict[str, Any]) -> TDTO:
        rs = self._execute(query, params).mappings().first()
        if not rs:
            raise NotFound('Evento Not Found')
        return self._dto(**rs)

    def get(self, id: str) -> TDTO:
        query: str = f"SELECT * FROM {self._get_ftbl_name()}"
        filter_args = {self._dto.__PK__: id}
        query = self._filter_query(query, filter_args)
        return self._get(query, filter_args)

    def _filter_query(self, query: str, filter_args: dict[str, str]) -> str:
        keys = list(filter_args.keys())
        if keys:
            query += f" WHERE {keys[0]} = :{keys[0]}"
            keys.pop(0)
            for key in keys:
                query += f" AND {key} = :{key}"
        return query

    def _get_first(self, query: str) -> Any | None:
        rs = self._execute(query)
        frow = rs.first()
        if not frow:
            return None
        return frow[0]

    def get_all(self) -> list[TDTO]:
        query: str = f"SELECT * FROM {self._get_ftbl_name()}"
        rs = self._execute(query)
        return [
            self._dto(**x)
            for x in rs.mappings().all()
        ]

    def list(self, pagi_args   : dict[str, str | int]
                 , filter_args : dict[str, str] = {}) -> list[TDTO]:
        query: str = f"SELECT * FROM {self._get_ftbl_name()}"
        return self._list(query, pagi_args, filter_args)

    def get_rows_total(self) -> int:
        query: str = f"SELECT count(*) FROM {self._get_ftbl_name()}"
        ret = self._get_first(query)
        if not ret:
            return -1
        return int(ret)
